import base64

from datalogue.clients.http import _HttpClient, Union, Optional, HttpMethod
from datalogue.clients.jobs import _JobsClient
from datalogue.clients.datastore_collections import _DatastoreCollectionClient
from datalogue.clients.credentials import _CredentialsClient
from datalogue.clients.datastore import _DatastoreClient
from datalogue.clients.organization import _OrganizationClient
from datalogue.clients.group import _GroupClient
from datalogue.clients.tag import _TagClient
from datalogue.clients.test_auth_client import _TestAuthenticationClient
from datalogue.clients.user import _UserClient
from datalogue.clients.ontology import _OntologyClient
from datalogue.clients.ml import _MLClient
from datalogue.clients.classifier import _ClassifierClient
from datalogue.clients.regex import _RegexClient
from datalogue.clients.pipeline import _PipelineClient
from datalogue.credentials import DtlCredentials, _get_ssl_verify_env
from datalogue.clients.authentication import _AuthenticationClient
from datalogue.errors import DtlError

from uuid import UUID
from urllib.parse import urlparse
import os

from datalogue.models.organization import User
from datalogue.models.version import Version

class LoginCredentials:
    def __init__(self, uri: str):
        self.uri = uri


class DtlTokenCredentials(LoginCredentials):
    def __init__(self, uri: str, user_id: Union[str, UUID], token: str):
        LoginCredentials.__init__(self, uri)
        self.user_id = user_id
        self.token = token

class Dtl:
    """
    Root class to be built to interact with all the services

    :param credentials: contains the information to connect
    """

    def __init__(self, credentials: LoginCredentials):
        self.uri = credentials.uri

        if isinstance(credentials, DtlCredentials):
            self.username = credentials.username
            self.http_client = _HttpClient(credentials.uri, credentials.authentication_name, credentials.verify_certificate)

            login_res = self.http_client.login(credentials.username, credentials.password)
            if isinstance(login_res, DtlError):
                raise login_res

        if isinstance(credentials, DtlTokenCredentials):
            user_id_with_token = f'{credentials.user_id}:{credentials.token}'
            encoded_bytes = base64.b64encode(user_id_with_token.encode("utf-8"))
            encoded_token = str(encoded_bytes, "utf-8")

            self.http_client = _HttpClient(credentials.uri, encoded_auth_header = f'TOKEN {encoded_token}')
            current_user = _UserClient(self.http_client).get_current_user()
            self.username = current_user.email

        self.group = _GroupClient(self.http_client)
        """Client to interact with the groups"""
        self.user = _UserClient(self.http_client)
        """Client to interact with the users"""
        self.organization = _OrganizationClient(self.http_client)
        """Client to interact with the organization part of the stack"""
        self.jobs = _JobsClient(self.http_client)
        """Client to interact with the jobs"""
        self.datastore_collection = _DatastoreCollectionClient(self.http_client)
        """Client to interact with the datastore collections"""
        self.datastore = _DatastoreClient(self.http_client)
        """Client to interact with the datastores"""
        self.credentials = _CredentialsClient(self.http_client)
        """Client to interact with credentials"""
        self.ontology = _OntologyClient(self.http_client)
        """Client to interact with ontologies"""
        self.ml = _MLClient(self.http_client)
        """Client to interact with ml"""
        self.classifier = _ClassifierClient(self.http_client, self.ontology)
        """Client to interact with classifiers"""
        self.regex = _RegexClient(self.http_client)
        """Client to interact with regexes"""
        self.tag = _TagClient(self.http_client)
        """Client to interact with tags"""
        self.pipeline = _PipelineClient(self.http_client)
        """Client to interact with the stream collections"""
        self.authentication = _AuthenticationClient(self.http_client)
        """Client to interact with the Authentication Schemes"""

    def __repr__(self):
        return f'Logged in {self.uri!r} with {self.username!r} account.'

    @staticmethod
    def signup(uri='', first_name='', last_name='', email='', password='', accept_terms=True,
               verify_certificate: Optional[bool] = None
               ) -> Union[DtlError, User]:
        """
        Perform signup of user
        :param uri: The target URI where the user data will be associated in
        :param accept_terms: Whether the user accept the following terms : https://www.datalogue.io/pages/terms-of-service
        :return: User object if successful, else return DtlError
        """

        if verify_certificate is None:
            verify_certificate = _get_ssl_verify_env()

        http_client = _HttpClient(uri, verify_certificate=verify_certificate)
        http_client.get_csrf()
        registered_user = http_client.signup(first_name, last_name, email, password, accept_terms)
        return registered_user

    def version(self) -> Union[DtlError, Version]:
        """
        Get version number of SDK, platform, and the platform's services'
        :return: Version object containing version number of SDK, platform, and the platform's services'
        """
        res = self.http_client.make_authed_request("/version", HttpMethod.GET)
        if isinstance(res, DtlError):
            return res
        return Version.from_payload(res)





from typing import Optional, Union, List, Dict, Tuple, Iterator

from itertools import chain
from typing import AnyStr, Callable
import statistics
import pandas as pd
import numpy as np
from uuid import UUID, uuid4
from pathlib import Path

from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.clients.ontology import _OntologyClient
from datalogue.models.ontology import Ontology, OntologyNode, DataRef, TrainingDataColumn
from datalogue.models.model import Model, ModelType, TrainingStatusType, OrderList, TrainingState, \
    training_status_type_from_str, Deployment, Section, SynthesizedSection, OriginalSection, SynthesisStats, \
    TrainingData
from datalogue.errors import DtlError
from datalogue.dtl_utils import _parse_list, _parse_uuid
from uuid import UUID
from datalogue.models.metrics import ArgoMetrics


class _SynthesisClient:
    """
    Client to interact with SynthesizedData
    """
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def _get_sections(self, endpoint: str, datastore_ids: List[Union[str, UUID]], page: int, size: int) -> \
            Union[DtlError, dict]:
        params = {
            "datastoreId": datastore_ids,
            "page": page,
            "size": size
        }
        res = self.http_client.make_authed_request(endpoint, HttpMethod.GET, params=params)
        if isinstance(res, DtlError):
            return res
        return res

    def get(self, datastore_ids: List[Union[str, UUID]], ontology_id: Union[str, UUID], page: int, page_size: int) -> \
            Union[DtlError, List[SynthesizedSection]]:
        """
        Retrieves all data that was generated by the run of the synthesize function. The parameters filter the list of section returned

        :param datastore_ids: list of datastores to be used to filter the section to be used as a baseline. If None is used, then it uses all the synthesize all the data annotated for a given ontology
        :param ontology_id: id of the ontology for which we are doing the synthesizing
        :param page: which page of data to be returned
        :param page_size: number of elements in the page
        :return: List of Sections

        """
        res = self._get_sections(f"/themis/sections/ontology/{ontology_id}/synthesized", datastore_ids, page, page_size)
        if isinstance(res, DtlError):
            return res
        return _parse_list(SynthesizedSection.from_payload)(res)

    def _original(self, datastore_ids: List[Union[str, UUID]], ontology_id: Union[str, UUID], page: int, size: int):
        res = self._get_sections(f"/themis/sections/ontology/{ontology_id}/original", datastore_ids, page, size)
        if isinstance(res, DtlError):
            return res
        return _parse_list(OriginalSection.from_payload)(res)

    def _count_original(self, datastore_ids: List[Union[str, UUID]], ontology_id: Union[str, UUID]):
        params = {
            "datastoreId": datastore_ids,
        }
        res = self.http_client.make_authed_request(f"/themis/sections/ontology/{ontology_id}/original/count", HttpMethod.GET, params=params)
        if isinstance(res, DtlError):
            return res
        if isinstance(res, int):
            return res
        return DtlError(f"{res} is expected to be int")

    def _create(self, ontology_id: Union[str, UUID], section: SynthesizedSection) -> Union[DtlError, None]:
        body = section._as_payload()
        return self.http_client.make_authed_request(f"/themis/sections/ontology/{ontology_id}/synthesized",
                                                       HttpMethod.POST, body=body)

    def _clean(self, datastore_ids: List[Union[str, UUID]],
               ontology_id: Union[str, UUID]) -> Union[DtlError, None]:
        params = {
            "datastoreId": datastore_ids,
        }
        return self.http_client.make_authed_request(f"/themis/sections/ontology/{ontology_id}/synthesized", HttpMethod.DELETE, params=params)

    def synthesize(self,
                   datastore_ids: List[Union[str, UUID]],
                   ontology_id: Union[str, UUID],
                   synthesis_algorithm: Callable[[OriginalSection], List[SynthesizedSection]],
                   overwrite_synthesis=False
                   ) -> Union[DtlError, SynthesisStats]:
        """
        Creates a synthesized training data set. That Data is generated from some original annotated training data.

        :param datastore_ids: list of datastores to be used to filter the section to be used as a baseline. If None is used, then it uses all the data annotated for a given ontology
        :param ontology_id: id of the ontology for which we are doing the synthesization
        :param synthesis_algorithm: function that is to be called on sections and synthesizes training data
        :param overwrite_synthesis: when set to true, the previously synthesized data is deleted and replaced by the one generated by this call. This flag takes into account the filters that are activated (ontology or datastore)
        :return: Statistics about the synthesis
        """
        if overwrite_synthesis:
            self._clean(datastore_ids, ontology_id)

        current_page = 1
        batch_size = 10
        to_process = self._count_original(datastore_ids, ontology_id)
        if isinstance(to_process, DtlError):
            return DtlError("Nothing to process")
        page = self._original(datastore_ids, ontology_id, current_page, batch_size)
        processed = 0
        created = []

        while not isinstance(page, DtlError) and len(page) > 0:
            prepared_to_send = [synthesis_algorithm(item) for item in page]
            to_create = chain.from_iterable(prepared_to_send)
            created_in_batch = 0
            for section in to_create:
                res = self._create(ontology_id, section)
                if isinstance(res, DtlError):
                    return res
                created_in_batch = created_in_batch + 1
            created.append(created_in_batch)
            processed = processed + len(page)
            current_page = current_page + 1
            print(f"{processed/to_process*100}%")
            if len(page) < batch_size:
                break
            page = self._original(datastore_ids, ontology_id, current_page, batch_size)
        print("Success!")
        if len(created) > 1:
            return SynthesisStats(sum(created), processed, statistics.mean(created), statistics.stdev(created))
        else:
            return SynthesisStats(sum(created), processed, statistics.mean(created), 0)


class TrainingCoach:

    @staticmethod
    def check(ontology: Ontology) -> pd.DataFrame:
        leaves = ontology.leaves()
        rows = []
        for leaf in leaves:
            is_without_data = leaf.total_training_count == 0
            unique_names = set(map(lambda x: x.datastore_name, leaf.training_data_columns))
            has_got_only_one_source = len(unique_names) < 2
            total_smaller_than_x = leaf.total_training_count < 2000
            rows.append([leaf.name, leaf.id, is_without_data, has_got_only_one_source, total_smaller_than_x])

        return pd.DataFrame(rows,
                            columns=['Class', 'Id', 'WARNING: No data', 'WARNING: <2 sources', 'WARNING: <2000 data points'])


class _MLBase:
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
    
    def get(self, model_id: UUID) -> Union[DtlError, TrainingState]:
        """""
        Get :class:`Trainings` based on the given model_id

        :param model_id:
        :return: Requested training if successful, else returns :class:`DtlError`
        """

        res = self.http_client.make_authed_request(f'/argus/trainings/{str(model_id)}',
                                                   HttpMethod.GET)
        if isinstance(res, DtlError):
            return DtlError(f"Something happened while trying to get training for {model_id}", res)

        return TrainingState._from_payload(res)
    
    def cancel(self, model_id: UUID) -> Union[DtlError, bool]:
        """
        Cancel training given a model_id

        :param model_id:
        :return: status of the training if successful, else returns :class:`DtlError`
        """

        res = self.http_client.make_authed_request(
            f'/argus/trainings/{str(model_id)}/cancel',
            HttpMethod.POST)

        if isinstance(res, DtlError):
            return res
        else:
            return True

    def get_model_url(self, model_id: UUID) -> Union[DtlError, Optional[str]]:
        """
        Return url of a deployment for a given model_id

        :param model_id:
        :return: url if there is an active deployment for the training, otherwise returns None
        """

        res = self.http_client.make_authed_request(f'/argus/deployments/model-url/{str(model_id)}', HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return res["url"]

    def get_model_history(self, ontology_id: UUID, limit: Optional[int] = 20) -> Union[DtlError, List[Model]]:
        """
        Get :class:`Training` history based on the given ontology_id
        
        :param ontology_id:
        :return: List of trainings if successful, else returns :class:`DtlError`
        """

        params = {}
        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")
            params["limit"] = limit

        res = self.http_client.make_authed_request(
            f'/yggy/ontology/{str(ontology_id)}/trainings/history',
            HttpMethod.GET, params=params)

        if isinstance(res, DtlError):
            return res
        else:
           return _parse_list(Model._from_payload)(res)


class _MLClient:
    """
    Client to interact with the ML
    """
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.cbc = _CBCClient(http_client)
        self.ner = _NERClient(http_client)
        self.deployment = _DeploymentClient(http_client)


class _CBCDataClient:
    """
    Client to interact with TrainingData
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def add(self, store_id: UUID, store_name: str, refs: List[DataRef], count: Optional[int] = None) -> Union[DtlError, List[UUID]]:
        """
        Attaches paths in the given `store_id` to the the nodes of the ontology.

        :param store_id: Id of the datastore that it going to be read
        :param store_name: Name of the datastore that it going to be read
        :param refs: List of data references to show which paths are going to be attached which ontology nodes
        :param count: Number of n first rows which we want to attach, if not passed we attach all rows
        :return: List of stream ids which are the jobs that is transferring data from datastore to Themis
        """
        stream_ids = []

        for dataRef in refs:
            for path in dataRef.path_list:

                # Creates a dataset per attachment as UI does.
                dataset_id = self.__create_dataset(store_name)
                if isinstance(dataset_id, DtlError):
                    return dataset_id

                stream_id = self.__transfer_data_from_datastore(store_id, dataset_id, dataRef.node_id, path, count)
                if isinstance(stream_id, DtlError):
                    return stream_id

                self.__update_node(dataRef.node_id, path, dataset_id, stream_id)
                stream_ids.append(stream_id)
        return stream_ids

    def __create_dataset(self, store_name: str) -> Union[DtlError, UUID]:
        payload = {
            "title": store_name,
            "tags": [],
            "label_map": {}
        }

        res = self.http_client.make_authed_request('/themis/dataset', HttpMethod.POST, body=payload)
        dataset_id = UUID(res.get("id"))

        if dataset_id is None:
            return DtlError("There is no dataset id in response!")
        return dataset_id

    def __transfer_data_from_datastore(self, store_id: UUID, dataset_id: UUID, node_id: UUID, path: List[str], count: Optional[int]) -> UUID:
        params = {
            'sourceId': str(store_id),
            'trainingDataId': str(dataset_id),
            'class': str(node_id),
            'path': path
        }
        if count is not None:
            params['count'] = count
        res = self.http_client.make_authed_request('/scout/run/training-data', HttpMethod.POST, params=params)

        if isinstance(res, DtlError):
            return res

        stream_id = UUID(res["streamId"])
        return stream_id

    def __update_node(self, node_id: UUID, path: List[str], dataset_id: UUID, stream_id: UUID) -> List[str]:
        #TODO We should remove this part when we move training data adding functionality out of Yggy
        entity_res = self.http_client.make_authed_request(f"/yggy/entity/{node_id}", HttpMethod.GET)
        training_data_list = entity_res.get("trainingDataInfo")
        if training_data_list is None:
            training_data_list = []

        training_data_list.append({
            "datasetId": str(dataset_id),
            "nodePath": path,
            "streamId": str(stream_id)
        })

        self.http_client.make_authed_request(f"/yggy/entity/{node_id}", HttpMethod.POST,
                                                body={"trainingDataInfo": training_data_list})

        dataset_ids = []
        for trainingData in training_data_list:
            dataset_ids.append(trainingData["datasetId"])

        return dataset_ids


class _CBCClient(_MLBase):
    """
    Client to interact with the CBC
    """  
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.data = _CBCDataClient(http_client)

    def get_deployments(self,
                        page: int = 1,
                        item_per_page: int = 25) -> Union[DtlError, List[TrainingState]]:
        """
        Retrieve a list of all models currently deployed.

        :param page:
        :param item_per_page:
        :return:
        """
        from_item = (page-1)*item_per_page
        params = {
            'from': from_item,
            'limit': item_per_page,
            'status': 'Successful',
            'model-type': str(ModelType.Cbc)
        }

        res = self.http_client.make_authed_request(
            f'/argus/deployments', HttpMethod.GET, params=params)
        if isinstance(res, DtlError):
            return res
        else:
            return _parse_list(Deployment._from_payload)(res)

    def deploy(self, model_id: UUID, ontology_id: UUID) -> Union[DtlError, UUID]:
        """
        Deploy :class:`Training` based on the given model_id and ontology_id
        :return: True if successful, else returns :class:`DtlError`
        """
        res = self.http_client.make_authed_request(
            f'/yggy/ontology/{str(ontology_id)}/trainings/{str(model_id)}/deploy',
            HttpMethod.POST, {})

        if isinstance(res, DtlError):
            return res
        else:
            deployment_id = UUID(res.get('deploymentId'))
            return deployment_id

    def train(self, ontology_id: UUID, force_run: bool = True) -> Union[DtlError, UUID, List]:
        """
        Starts training for a given ontology_id

        :param ontology_id: id of the ontology to be used for training
        :param force_run: allows to run training even when there are less data than should be
        :return: Either a :class:`DtlError` if any error occurs or :class:`True` if training is requested successfully.
        """

        params = dict()
        params["model-type"] = ModelType.Cbc
        params["force-train"] = force_run

        res = self.http_client.make_authed_request(f"/yggy/ontology/{ontology_id}/train", HttpMethod.POST, params=params)
        if isinstance(res, DtlError):
            return res
        if isinstance(res, List):
            # TODO we should handle response with given content: [{'TagsLessThanLimit': {'currentTotalTags': X, 'recommendedTags': Y}}]
            return res
        else:
            model_id = UUID(res.get('trainingId'))
            return model_id

    def get_models(self, ontology_id: UUID, status_type: Optional[TrainingStatusType] = None, sort: Optional[OrderList] = OrderList.desc, limit: Optional[int] = 20) -> Union[DtlError, List[TrainingState]]:
        """
        Get :class:`Trainings` based on the given ontology_id
        
        :param ontology_id:
        :return: List of trainings if successful, else returns :class:`DtlError`
        """

        params = {}

        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")
            params["limit"] = limit

        params["model-type"] = ModelType.Cbc

        if status_type is not None:
            params["status-type"] = status_type.value

        if sort is not None:
            params["sort"] = sort.value

        res = self.http_client.make_authed_request(
            f'/yggy/ontology/{str(ontology_id)}/trainings', HttpMethod.GET, params=params)

        if isinstance(res, DtlError):
            return res
        else:
            return _parse_list(TrainingState._from_payload)(res)


    def upload(self, ontology_id: UUID, name: str, argo_work_dir: Path, experiment_name: str, class_map_path: str, model_id: Optional[UUID] = None) -> Optional[bool]:
        """	
        Processes and pushes via the Argus api to the defined	
        ontology.	
        :param ontology_id: UUID the ontology you are publishing to	
        :param name: str who is pushing this data	
        :param argo_work_dir: str Path to the template's work directory (i.e. nightshade/work/cbc)	
        :param experiment_name: str Name of the experiment to upload	
        :param class_map_path: str	
        :return: Error if it fails or True otherwise	
        """
        if model_id is None:
            model_id = str(uuid4())

        M = ArgoMetrics(str(argo_work_dir), experiment_name, class_map_path)
        endpoint = f'/yggy/ontology/{ontology_id}/trainings/upload-metrics'
        params = {
            "training-id": f"{model_id}",
            "user-firstname": f"{name}",
            "user-lastname": "SDK",
        }

        res = self.http_client.make_authed_request(endpoint, HttpMethod.PUT, M.payload, params)

        if isinstance(res, DtlError):
            return DtlError("Could not publish request.", res)

        return True

    def coach(self, ontology_id: Union[UUID, str]) -> Union[DtlError, pd.DataFrame]:
        """
        Receive machine learning coaching instructions based on the training data attached to an ontology
        For each class, see if the data attached is insufficient in volume or has been sourced from unvaried data, potentially leading to overfit
        :param ontology_id: id of the ontology on which to receive machine learning coaching
        :return: dataFrame containing coaching if successful, or DtlError of failed
        """

        ontology_client = _OntologyClient(self.http_client)
        uuid = _parse_uuid(ontology_id)
        if isinstance(uuid, DtlError):
            return uuid
        ontology = ontology_client.get(uuid)
        if isinstance(ontology, DtlError):
            return ontology
        return TrainingCoach.check(ontology)


class _NERDataClient:
    """
    Client to interact with TrainingData
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.synthesis = _SynthesisClient(http_client)

    def delete_section(self, datastore_id: Union[str, UUID], path: List[str]) -> Union[DtlError, None]:
        """
        Deletes a section and all associated annotations by a given section_id
        :param section_id: id of the section to be deleted
        :return: Either a :class:`DtlError` if any error occurs or None if deletion is successful.
        """
        datastore_id = _parse_uuid(datastore_id, f"The supplied datastore_id: {datastore_id} is not a valid UUID")
        if isinstance(datastore_id, DtlError):
            return datastore_id
        res = self.http_client.make_authed_request(f'/themis/sections', HttpMethod.DELETE, body={
            "datastore_id": str(datastore_id),
            "path": path
        })
        if isinstance(res, DtlError):
            return res
        else:
            return None

    def get_annotations(self, ontology_id: Union[str, UUID],
                        datastore_ids: List[Union[str, UUID]],
                        items_per_page: int = 25,
                        page_after_id: Union[str, UUID, None] = None) -> Union[DtlError, Tuple[List[Section], Optional[UUID]]]:
        """
        Retrieve a list of annotated sections tagged from an ontology, with full text, all associated annotations of that section, and other statistics.
        :param ontology_id: id of the ontology to retrieve annotated sections from
        :param datastore_ids: can be used to retrieve only sections imported from specified datastores, if None, will return all annotated sections
        :param items_per_page: the number of AnnotatedSections to return
        :param page_after_id: the id of an AnnotatedSection in sequence, after which the returned page begins
        :return: a list of AnnotatedSections with value which should be used as page_after_id in next call if successful,
        or DtlError if failed

        """
        # validate ontology_id
        ontology_id = _parse_uuid(ontology_id, f"The supplied ontology_id: {ontology_id} is not a valid UUID")
        if isinstance(ontology_id, DtlError):
            return ontology_id

        # validate page_after_id
        if page_after_id is not None:
            page_after_id = _parse_uuid(
                ontology_id, f"The supplied page_after_id: {page_after_id} is not a valid UUID")
            if isinstance(page_after_id, DtlError):
                return page_after_id
        after_query = f'&after={page_after_id}' if page_after_id is not None else ''
        uri = f"/themis/ontology/{ontology_id}/export-annotations-json?size={items_per_page}{after_query}"

        # validate datastore_ids
        datastore_ids = list(map(lambda datastore_id: _parse_uuid(
            datastore_id, f"The supplied datastoreId: {datastore_id} is not a valid UUID"), datastore_ids))
        for datastore_id in datastore_ids:
            if isinstance(datastore_id, DtlError):
                return datastore_id

        body = list(map(lambda datstore_id: str(datstore_id), datastore_ids))
        res = self.http_client.execute_authed_request(uri, HttpMethod.POST, body)

        if isinstance(res, DtlError):
            return res
        sections_statistics = _parse_list(Section.from_payload)(res.json())
        if isinstance(sections_statistics, DtlError):
            return sections_statistics
        next_token_raw = res.headers.get("X-Next-Token", None)
        next_token = _parse_uuid(next_token_raw) if next_token_raw is not None else None
        if isinstance(next_token, DtlError):
            return next_token
        return sections_statistics, next_token


class _NERClient(_MLBase):
    """
    Client to interact with the NER
    """
    
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.data = _NERDataClient(http_client)

    def get_deployments(self,
                        page: int = 1,
                        item_per_page: int = 25) -> Union[DtlError, List[TrainingState]]:
        """
        Retrieve a list of all models currently deployed.
        :param page:
        :param item_per_page:
        :return:
        """
        from_item = (page-1)*item_per_page
        params = {
            'from': from_item,
            'limit': item_per_page,
            'status': 'Successful',
            'model-type': str(ModelType.Ner)
        }
        
        res = self.http_client.make_authed_request(
            f'/argus/deployments', HttpMethod.GET, params=params)
        if isinstance(res, DtlError):
            return res
        else:
            return _parse_list(Deployment._from_payload)(res)

    def deploy(self, model_id: UUID, ontology_id: UUID) -> Union[DtlError, UUID]:
        """
        Deploy :class:`Training` based on the given model_id and ontology_id
        :return: True if successful, else returns :class:`DtlError`
        """
        res = self.http_client.make_authed_request(
            f'/yggy/ontology/{str(ontology_id)}/trainings/{str(model_id)}/deploy',
            HttpMethod.POST, {})

        if isinstance(res, DtlError):
            return res
        else:
            deployment_id = UUID(res.get('deploymentId'))
            return deployment_id

    def train(self, ontology_id: UUID, training_data: TrainingData = TrainingData.Original, force_run: bool = True) -> Union[DtlError, UUID, List]:
        """
        Starts training for a given ontology_id

        :param ontology_id: id of the ontology to be used for training
        :param training_data: type of training data to be used to train the model: TrainingData.Original or TrainingData.Synthesized
        :param force_run: allows to run training even when there are less data than should be
        :return: Either a :class:`DtlError` if any error occurs or :class:`True` if training is requested successfully.
        """

        params = dict()
        params["model-type"] = ModelType.Ner
        params["data-origin"] = training_data.value
        params["force-train"] = force_run

        res = self.http_client.make_authed_request(
            f"/yggy/ontology/{ontology_id}/train", HttpMethod.POST, params=params)
        
        if isinstance(res, DtlError):
            return res
        if isinstance(res, List):
            # TODO we should handle response with given content: [{'TagsLessThanLimit': {'currentTotalTags': X, 'recommendedTags': Y}}]
            return res
        else:
            model_id = UUID(res.get('trainingId'))
            return model_id

    def get_models(self, ontology_id: UUID, status_type: Optional[TrainingStatusType] = None, sort: Optional[OrderList] = OrderList.desc, limit: Optional[int] = 20) -> Union[DtlError, List[TrainingState]]:
        """
        Get :class:`Trainings` based on the given ontology_id
        
        :param ontology_id:
        :return: List of trainings if successful, else returns :class:`DtlError`
        """

        params = {}

        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")
            params["limit"] = limit

        params["model-type"] = ModelType.Ner

        if status_type is not None:
            params["status-type"] = status_type.value

        if sort is not None:
            params["sort"] = sort.value

        res = self.http_client.make_authed_request(
            f'/yggy/ontology/{str(ontology_id)}/trainings', HttpMethod.GET, params=params)

        if isinstance(res, DtlError):
            return res
        else:
            return _parse_list(TrainingState._from_payload)(res)

class _DeploymentClient:
    """
    Client to interact with the Deployments
    """
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def get(self, deployment_id: UUID):
        res = self.http_client.make_authed_request(f'/argus/deployments/{str(deployment_id)}', HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return Deployment._from_payload(res)



from typing import Optional
from datalogue.models.datastore import DatastoreDef
from datalogue.models.credentials import Credentials


class Datastore:
    def __init__(self, definition: DatastoreDef, credentials: Credentials):
        self.definition = definition
        self.credentials = credentials

    def _as_payload(self) -> dict:
        base = self.definition._as_payload().copy()
        if self.credentials is not None:
            base.update(self.credentials._as_payload())
        return base


from datalogue.models.transformations.commons import Transformation, DataType
from datalogue.dtl_utils import _parse_string_list, SerializableStringEnum
from datalogue.errors import _enum_parse_error, DtlError, _property_not_found, _invalid_property_type
from typing import List, Union


class Period:

    def __init__(self, identifier: str, value: int, unit: str):
        self.identifier = identifier
        self.value = value
        self.unit = unit

    def __eq__(self, other: 'Period'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"Period(identifier: {self.identifier}, value: {self.value}, unit: {self.unit})"

    def _as_payload(self) -> dict:
        return {
            "identifier": self.identifier,
            "value": self.value,
            "unit": self.unit
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Period']:
        identifier = json.get("identifier")
        if identifier is None:
            return _property_not_found("identifier", json)

        value = json.get("value")
        if value is  None:
            return _property_not_found("value", json)

        unit = json.get("unit")
        if unit is  None:
            return _property_not_found("unit", json)

        return Period(identifier, value, unit)


class Format:

    def __init__(self, start: str, end: str, separator: str):
        self.start = start
        self.end = end
        self.separator = separator

    def __eq__(self, other: 'Format'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"Format(start: {self.start}, end: {self.end}, separator: {self.separator})"

    def _as_payload(self) -> dict:
        return {
            "start": self.start,
            "separator": self.separator,
            "end": self.end
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Format']:
        start = json.get("start")
        if start is None:
            return _property_not_found("start", json)

        separator = json.get("separator")
        if separator is  None:
            return _property_not_found("separator", json)

        end = json.get("end")
        if end is  None:
            return _property_not_found("end", json)

        return Format(start, end, separator)


class Output:

    def __init__(self, start_label: str, end_label: str, format: str):
        self.start_label = start_label
        self.end_label = end_label
        self.format = format

    def __eq__(self, other: 'Output'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"Output(start: {self.start_label}, end: {self.end_label}, separator: {self.format})"

    def _as_payload(self) -> dict:
        return {
            "startLabel": self.start_label,
            "endLabel": self.end_label,
            "format": self.format
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Output']:
        start_label = json.get("startLabel")
        if start_label is None:
            return _property_not_found("startLabel", json)

        end_label = json.get("endLabel")
        if end_label is  None:
            return _property_not_found("endLabel", json)

        format = json.get("format")
        if format is  None:
            return _property_not_found("format", json)

        return Output(start_label, end_label, format)



from typing import Union

from datalogue.errors import DtlError, _enum_parse_error
from datalogue.dtl_utils import SerializableStringEnum


class Scope(SerializableStringEnum):
    User = "User"
    Group = "Group"
    Organization = "Organization"

    @staticmethod
    def parse_error(s: str) -> str:
        return DtlError(_enum_parse_error("share target", s))

    @staticmethod
    def from_string(string: str) -> Union[DtlError, 'Scope']:
        return SerializableStringEnum.from_str(Scope)(string)



from datalogue.models.transformations.commons import Transformation
from datalogue.dtl_utils import _parse_string_list, _parse_list
from datalogue.errors import DtlError, _property_not_found
from typing import List, Union


class ConcatenateAtPaths(Transformation):
    """
    Concatenates the data that reside at specified sibling nodes,
    This operation is performed[ in-place: specified nodes are collapsed to one node containing the result of the concatenation as the value,
    and the same label as in the original.
    """
    type_str = "ConcatenateAtPaths"

    def __init__(self, paths: List[List[str]]):
        """
        Paths --> List of Paths where the data to concatenate resides
        """
        Transformation.__init__(self, ConcatenateAtPaths.type_str)
        self.paths = paths

    def __eq__(self, other: 'ConcatenateAtPaths'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ConcatenateAtPaths({self.paths})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["paths"] = self.paths
        return base

    @staticmethod
    def _list_paths_from_payload(value: dict) -> Union[List[List[str]], DtlError]:
        def is_list_of_list():
            return isinstance(value, List) and all(isinstance(item, List) for item in value)
        if value is not None and is_list_of_list():
            return _parse_list(_parse_string_list)(value)

        return DtlError(f"Could not parse operand: {value}")

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'ConcatenateAtPaths']:
        paths = json.get("paths")
        if paths is None:
            return _property_not_found("paths", json)
        paths = ConcatenateAtPaths._list_paths_from_payload(paths)
        if isinstance(paths, DtlError):
            return paths

        return ConcatenateAtPaths(paths)



import datetime
from abc import ABC, abstractmethod
from typing import List, Union
from datalogue.errors import DtlError
from datalogue.dtl_utils import _parse_string_list

class ShardIterator(ABC):
    """
    Abstract class containing the five subclasses:
      TrimHorizon, for beginning at the beginning of the stream
      Latest, for beginning at the most recent record
      AtTimestamp, for beginning at a specified time
      AtSequenceNumber, for beginning at a specified sequence number
      AfterSequenceNumber, for beginning just after a specified sequence number
    """
    type_field = "type"

    type_str = ""
    def __init__(self, type: str):
        self.type = type
        super().__init__()

    def _base_payload(self) -> dict:
        base = [(ShardIterator.type_field, self.type_str)]
        return dict(base)

    @abstractmethod
    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        pass

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, 'ShardIterator']:
        type_field = d.get(ShardIterator.type_field)
        if not isinstance(type_field, str):
            return DtlError("The type is missing from JSON")
        elif type_field == TrimHorizon.type_str:
            return TrimHorizon()
        elif type_field == Latest.type_str:
            return Latest()
        elif type_field == AtSequenceNumber.type_str:
            return AtSequenceNumber._from_payload(d)
        elif type_field == AfterSequenceNumber.type_str:
            return AfterSequenceNumber._from_payload(d)
        elif type_field == AtTimestamp.type_str:
            return AtTimestamp._from_payload(d)
        else:
            return DtlError("The object %s is not a ShardIterator definition" % str(d))


class TrimHorizon(ShardIterator):
    """
      Begins reading stream from the beginning
    """

    type_str = "TrimHorizon"

    def __init__(self):
        super().__init__(TrimHorizon.type_str)

    def __repr__(self):
        return f'{self.__class__.__name__}'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        return base

class Latest(ShardIterator):
    """
      Begins reading stream from the beginning
    """

    type_str = "Latest"

    def __init__(self):
        super().__init__(Latest.type_str)

    def __repr__(self):
        return f'{self.__class__.__name__}'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        return base

class AtSequenceNumber(ShardIterator):
    """
      Begins reading stream starting from element of specified sequence number
    """

    type_str = "AtSequenceNumber"

    def __init__(self, sequence_number: str):
        self.sequence_number = sequence_number
        super().__init__(AtSequenceNumber.type_str)

    def __repr__(self):
        return f'{self.__class__.__name__}(sequence_number: {self.sequence_number})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base['sequenceNumber'] = self.sequence_number
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, 'AtSequenceNumber']:
        sequence_number = d.get("sequenceNumber")
        if sequence_number is None:
            return DtlError("The `sequenceNumber` is missing from JSON")
        return AtSequenceNumber(sequence_number)


class AfterSequenceNumber(ShardIterator):
    """
      Begins reading stream starting immediately after element of specified sequence number
    """

    type_str = "AfterSequenceNumber"

    def __init__(self, sequence_number: str):
        self.sequence_number = sequence_number
        super().__init__(AfterSequenceNumber.type_str)

    def __repr__(self):
        return f'{self.__class__.__name__}(sequence_number: {self.sequence_number})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base['sequenceNumber'] = self.sequence_number
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, 'AfterSequenceNumber']:
        sequence_number = d.get("sequenceNumber")
        if sequence_number is None:
            return DtlError("The `sequenceNumber` is missing from JSON")
        return AfterSequenceNumber(sequence_number)

class AtTimestamp(ShardIterator):
    """
      Begins reading stream from specified timestamp
    """
    type_str = "AtTimestamp"

    def __init__(self, timestamp: int):
        self.timestamp = timestamp
        super().__init__(AtTimestamp.type_str)

    def __repr__(self):
        return f'{self.__class__.__name__}(timestamp: {self.timestamp})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = self._base_payload()
        base['timestamp'] = self.timestamp
        return base

    @staticmethod
    def _from_payload(d: dict) -> Union[DtlError, 'AtTimestamp']:
        timestamp = d.get("timestamp")
        if timestamp is None:
            return DtlError("The `timestamp` is missing from JSON")
        timestamp = datetime.datetime.fromtimestamp(timestamp, tz = datetime.timezone.utc)
        if not isinstance(timestamp, datetime.datetime):
            return DtlError("The `timestamp` parsed from JSON is not a datetime format.")

        return AtTimestamp(timestamp)

class ShardAttributes(object):
    """
    Configures the shard attributes for the streaming source
    :param partition_keys can be used to specify to read from a specific partition key or partition keys. If unspecified, reads from all partition keys (consequently, all shards) of the stream
    :param shard_iterator: iterator type used for this source, as a subclass of ShardIterator
    :param refresh_interval: time between requests for data loads, in milliseconds
    :param limit: maximum number of transactions ingested from stream per second (for Kinesis, max supported is 5 and each transaction can contain 10k records)
    """

    def __init__(self, partition_keys: List[str] = [], shard_iterator: ShardIterator = TrimHorizon(),
                 refresh_interval: float = 1, limit: int = 5):
        if len(partition_keys) == 0 and (isinstance(shard_iterator, AtSequenceNumber) or isinstance(shard_iterator, AfterSequenceNumber)):
            raise DtlError("Please input the shard_id of the shard referenced by the specified sequence_number")
        self.partition_keys = partition_keys
        self.shard_iterator = shard_iterator
        self.refresh_interval = refresh_interval
        self.limit = limit

    def __repr__(self):
        return f'{self.__class__.__name__}(partition_keys: {self.partition_keys}, shard_iterator: {self.shard_iterator}, ' \
               f'refresh_interval: {self.refresh_interval}, limit: {self.limit})'


    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        json = {}
        json['partitionKeys'] = self.partition_keys
        json['shardIterator'] = self.shard_iterator._as_payload()
        json['refreshInterval'] = self.refresh_interval
        json['limit'] = self.limit
        return json

    @staticmethod
    def _from_payload(shard_attributes: dict) -> Union[DtlError, 'ShardAttributes']:
        partition_keys = _parse_string_list(shard_attributes['partitionKeys'])
        shard_iterator = ShardIterator._from_payload(shard_attributes['shardIterator'])
        refresh_interval = shard_attributes['refreshInterval']
        limit = shard_attributes['limit']
        return ShardAttributes(partition_keys, shard_iterator, refresh_interval, limit)



from typing import List, Union, Optional, Dict

from datalogue.models.permission import Permission, Scope, SharePermission, ObjectType, UnsharePermission
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.pipeline import *
from datalogue.models.pipeline_behavior import *
from datalogue.models.job import _job_from_payload, Job
from datalogue.errors import DtlError, _invalid_pagination_params
from uuid import UUID, uuid4
from datetime import datetime, timedelta
from dateutil.tz import UTC
from datalogue.dtl_utils import _parse_list, ResponseStream
from pyarrow import csv, Table

from datalogue.models.pipeline_template import PipelineTemplate


class _PipelineClient:
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"
        self.template = _TemplateClient(http_client)

    def create(self, name: str, pipeline_def: PipelineDef, is_resilient: bool = False) -> Union[DtlError, Pipeline]:
        """
        Create :class:`PipelineEntity` object given name and pipeline object
        
        :param is_resilient: True if the job of this pipeline will be retried
        :param name: String, a name for your pipeline
        :param pipeline_def: PipelineDef, is the definition of the Pipeline
        :return: :class:`Pipeline` object, else returns :class:`DtlError`
        """
        payload = {
            "name": str(name),
            "definition": pipeline_def._as_payload(),
            "resilient": is_resilient
        }

        res = self.http_client.make_authed_request(
            self.service_uri + '/streams',
            HttpMethod.POST,
            payload)

        if isinstance(res, DtlError):
            return DtlError("Encountered an error while creating the pipeline", res)

        return Pipeline._from_payload(res)

    def get(self, pipeline_id: Union[str, UUID]) -> Union[DtlError, Pipeline]:
        """
        Get :class:`Pipeline` object given pipeline_id
        :return: :class:`Pipeline` object, else returns :class:`DtlError`
        """
        uri = f'{self.service_uri}/streams/{str(pipeline_id)}'
        res = self.http_client.execute_authed_request(uri, HttpMethod.GET)

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            return Pipeline._from_payload(res.json())
        if res.status_code == 400:
            payload = res.json()
            return DtlError(
                f'Encountered an error while getting the pipeline! pipeline_id: {pipeline_id}, reason: {payload["error"]}')
        else:
            return self.http_client._handle_status_code(uri, res)

    def update(
            self,
            pipeline_id: Union[str, UUID],
            name: Optional[str] = None,
            pipeline_def: Optional[PipelineDef] = None,
            is_resilient: Optional[bool] = None
    ) -> Union[DtlError, Pipeline]:
        """
        Updates the Pipeline as specified.

        :param pipeline_id: id of the pipeline to update
        :param name: Updated name of the pipeline (Optional)
        :param pipeline_def: Updated definition of the pipeline (Optional)
        :return: :class:`Pipeline` object with updated values, else returns :class:`DtlError`
        """
        payload = {}

        if name is not None:
            payload["name"] = str(name)

        if pipeline_def is not None:
            payload["definition"] = pipeline_def._as_payload()

        if is_resilient is not None:
            payload["resilient"] = is_resilient

        res = self.http_client.make_authed_request(
            f'{self.service_uri}/streams/{str(pipeline_id)}',
            HttpMethod.POST,
            payload)

        if isinstance(res, DtlError):
            return DtlError(f'Encountered an error while updating the pipeline for {pipeline_id}', res)

        return Pipeline._from_payload(res)

    def delete(self, pipeline_id: Union[str, UUID]) -> Union[DtlError, bool]:
        """
        Delete a pipeline.

        :param pipeline_id: id of the pipeline to be deleted
        :return: True if successful, or DtlError if failed
        """
        uri = f'{self.service_uri}/streams/{str(pipeline_id)}'
        res = self.http_client.execute_authed_request(uri, HttpMethod.DELETE)
        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            return True
        elif res.status_code == 400:
            payload = res.json()
            return DtlError(
                f'Encountered an error while deleting the pipeline! pipeline_id: {pipeline_id}, reason: {payload["error"]}')
        else:
            return self.http_client._handle_status_code(uri, res)

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, Pipeline]:
        """
        List all pipelines.

        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all pipelines if successful, or DtlError if failed
        """
        params = {
            "page": str(page),
            "size": str(item_per_page)
        }

        res = self.http_client.make_authed_request(
            f'{self.service_uri}/streams',
            HttpMethod.GET,
            params=params)

        if isinstance(res, DtlError):
            return DtlError(f'Encountered an error while listing pipelines', res)

        return _parse_list(Pipeline._from_payload)(res)

    def orchestrate(self, pipeline_id: Union[str, UUID], wait_on: List[Union[str, UUID]] = []) -> Union[DtlError, Job]:
        """
        Orchestrate the running of a pipeline to trigger after the success of one or more separate jobs

        :pipeline_id id of the pipeline to be run
        :wait_on id(s) of jobs which need to successfully finish before the specified pipeline will run.
        Default value is empty string which means if user does not define, there will be no dependencies
        :return: Returns Job object or error
        """
        params = {
            "wait_on": wait_on
        }

        res = self.http_client.make_authed_request(
            f'{self.service_uri}/streams/{pipeline_id}/orchestrate',
            HttpMethod.POST,
            params=params)

        if isinstance(res, DtlError):
            return res

        jobs = _parse_list(_job_from_payload)(res)
        if isinstance(jobs, DtlError):
            return jobs
        elif len(jobs) == 0:
            return DtlError("Encountered an issue! There should be at least one job!")
        else:
            # API returns main jobs and additional jobs per target to be analyzed once this stream is completed.
            # In SDK, we're returning only the main job to the user
            return jobs[0]

    def schedule(self, pipeline_id: Union[str, UUID], run_date: datetime,
                 frequency: Optional[timedelta] = None,
                 abort_if_overlong: bool = False) -> Union[DtlError, Job]:
        """
        Schedules the pipeline to run at the given date

        :param pipeline_id: UUID id of the pipeline to schedule
        :param date: datetime for scheduling
        :param frequency: interval value (e.g float) with period (e.g hours) after which a job will be repeated
        :param abort_if_overlong: if True, avoids running a job cycle if previous cycle is running. (False by default)
        :return: Returns Job object or error
        """
        uri = f'{self.service_uri}/streams/{pipeline_id}/schedule'
        payload = {
            "runDate": run_date.astimezone(UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
            "abortIfOverlong": abort_if_overlong
        }
        if frequency is not None:
            payload.update({
                "intervalInSec": frequency.seconds
            })
        res = self.http_client.execute_authed_request(uri, HttpMethod.POST, payload)

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            jobs = _parse_list(_job_from_payload)(res.json())
            if isinstance(jobs, DtlError):
                return jobs
            elif len(jobs) == 0:
                return DtlError("Encountered an issue! There should be at least one job!")
            else:
                return jobs[0]
        elif res.status_code == 400:
            payload = res.json()
            return DtlError(
                f'Encountered an error while scheduling the pipeline! pipeline_id: {pipeline_id}, reason: {payload["error"]}')
        else:
            return self.http_client._handle_status_code(uri, res)

    def run(self, pipeline_id: Union[str, UUID]) -> Union[DtlError, Job]:
        """
        Runs the pipeline right now

        :param pipeline_id: id of the pipeline to run
        :return: Returns Job object or error
        """

        uri = f'{self.service_uri}/streams/{pipeline_id}/run'

        res = self.http_client.execute_authed_request(uri, HttpMethod.POST)

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 200:
            jobs = _parse_list(_job_from_payload)(res.json())
            if isinstance(jobs, DtlError):
                return jobs
            elif len(jobs) == 0:
                return DtlError("Encountered an issue! There should be at least one job!")
            else:
                return jobs[0]
        elif res.status_code == 400:
            payload = res.json()
            return DtlError(
                f'Encountered an error while scheduling the pipeline! pipeline_id: {pipeline_id}, reason: {payload["error"]}')
        else:
            return self.http_client._handle_status_code(uri, res)

    #TODO we need to use sth else than target_id to differentiate definitions
    def preview(self, target_id: Union[str, UUID], pipeline_def: PipelineDef, size: int = 10) -> Union[DtlError, Table]:
        """
        Returns the result of the execution of a particular definition in the pipeline
        
        :param target_id: id of the target datastore in the pipeline to retrieve the preview of
        :param pipeline_def: a pipeline definition containing the source, transformations, and some number of targets, one of which to preview
        :param size: the size of the output table, in records
        :return: pyarrow.Table if successful, DtlError otherwise
        """
        params = {
            "size": size,
            "file-format": "Csv",
            "sample-store-id": str(target_id)
        }

        res = self.http_client.make_authed_request(
            f'{self.service_uri}/streams/samples',
            HttpMethod.POST,
            body = pipeline_def._as_payload(),
            params = params,
            stream = True)

        if isinstance(res, DtlError):
            return res

        response_stream = ResponseStream(res.iter_content(1024))
        return csv.read_csv(response_stream)

    def templatize(self, pipeline_id: Union[str, UUID]) -> Union[DtlError, PipelineTemplate]:
        """
        Generate a PipelineTemplate with logic matching that of the specified pipeline,
        and with an auto-generated name.
        Once generated, the PipelineTemplate can be tweaked with any automatic or user
        defined-variables as desired, or saved directly to the backend via
        dtl.pipeline.template.create()

        :param pipeline_id: id of the pipeline for which a template is to be generated
        :return: PipelineTemplate object with auto-generated name if successful,
        DtlError otherwise
        """
        uri = f'{self.service_uri}/streams/{pipeline_id}/templatize'

        res = self.http_client.make_authed_request(uri, HttpMethod.POST)

        if isinstance(res, DtlError):
            return res

        else:
            pipeline_name = res.get("streamName")
            template_name = f"Pipeline Template from: {pipeline_name} ({pipeline_id})"
            template_yaml = res.get("templateYaml")
            template_obj = PipelineTemplate(template_name, template_yaml)
            return template_obj


class _TemplateClient:
    """
    Client to interact with Pipeline Templates
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def create(self,
               pipeline_template: PipelineTemplate,
               validate: bool = True) -> Union[DtlError, PipelineTemplate]:
        """
        Create and validate the structure of a pipeline template

        :param pipeline_template: a local pipeline template, to be created
        and saved on the platform
        :param validate: can be set to False to opt-out of template
        validation upon creation. Otherwise, the argument of
        dummy_input_map in the PipelineTemplate object will be used to
        complete the template
        :return: Returns created PipelineTemplate if successful, or DtlError if failed
        """

        if validate is False:
            pipeline_template.dummy_input_map = None

        if validate is True and pipeline_template.dummy_input_map is None:
            return DtlError("dummy_input_map must have key -> value if 'validate' argument is True")

        res = self.http_client.make_authed_request(
            f"{self.service_uri}/templates",
            HttpMethod.POST,
            pipeline_template._as_payload())

        if isinstance(res, DtlError):
            return res

        return PipelineTemplate._from_payload(res)

    def update(self, id: Union[str, UUID],
               name: Optional[str] = None,
               template: Optional[str] = None,
               dummy_input_map: Optional[Dict[str, str]] = None,
               validate: bool = True) -> Union[DtlError, PipelineTemplate]:
        """
        Update a template's name and/or template by its ID.
        :param id: id of the template to be updated
        :param name : optional, new name of the template
        :param template: optional, new template to update the existing template
        :param dummy_input_map:  an example of data that will be used to
        validate this template upon creation, unless disabled upon the
        creation step
        :param validate: can be set to False to opt-out of template
        validation upon update. Otherwise, the argument of
        dummy_input_map will be used to
        complete the template
        :return: Returns an updated template if successful, ors DtlError if failed
        """
        req_body = {}
        if name is not None:
            req_body['name'] = name
        if template is not None:
            req_body['template'] = template
        if dummy_input_map is not None:
            req_body['dummy_input_map'] = dummy_input_map

        if validate is False:
            req_body['dummy_input_map'] = None
        if validate is True and dummy_input_map is None:
            return DtlError("dummy_input_map must have key -> value if 'validate' argument is True")
        if dummy_input_map is not None and template is None:
            return DtlError("template must be supplied for dummy_input_map validation")

        res = self.http_client.make_authed_request(
            f"{self.service_uri}/templates/{str(id)}",
            HttpMethod.PUT,
            req_body)

        if isinstance(res, DtlError):
            return res

        return PipelineTemplate._from_payload(res)

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[PipelineTemplate]]:
        """
        List all templates.

        :param page: page to be retrieved
        :param item_per_page: number of templates to be put in a page
        :return: Returns a List of all the available templates if successful, or DtlError if failed
        """
        req_params = {}
        if isinstance(page, int) and page > 0:
            req_params['page'] = page
        else:
            return _invalid_pagination_params("page")
        if isinstance(item_per_page, int) and item_per_page > 0:
            req_params['size'] = item_per_page
        else:
            return _invalid_pagination_params("item_per_page")
        res = self.http_client.make_authed_request(
            f"{self.service_uri}/templates",
            HttpMethod.GET,
            params=req_params)
        if isinstance(res, DtlError):
            return res

        return _parse_list(PipelineTemplate._from_payload)(res)

    def get(self, id: Union[str, UUID]) -> Union[DtlError, PipelineTemplate]:
        """
        Retrieve a template by its ID.
        :param id: the id of the template to be retrieved locally
        :return: Returns a template if successful, or a DtlError if failed
        """
        res = self.http_client.make_authed_request(
            f"{self.service_uri}/templates/{str(id)}",
            HttpMethod.GET)
        if isinstance(res, DtlError):
            return res

        return PipelineTemplate._from_payload(res)

    def delete(self, id: Union[str, UUID]) -> Union[DtlError, bool]:
        """
        Delete a template by its ID.
        :param id: id of the template to be deleted
        :return: True if successful, or DtlError if failed
        """
        res = self.http_client.execute_authed_request(
            f"{self.service_uri}/templates/{str(id)}",
            HttpMethod.DELETE)

        if isinstance(res, DtlError):
            return res
        elif res.status_code == 204:
            return True
        else:
            return DtlError(res.text)

    def share(self, id: UUID, target_id: UUID, target_type: Scope, permission: Permission) -> Union[
        SharePermission, DtlError]:
        """
        Share a pipeline template.
        :param id: id of the template that is to be shared
        :param target_type: Scope (`User`, `Group` or `Organization`) with whom you want to share the template. It can be User, Group or Organization.
        :param target_id: id of the User, Group or Organization you want to share with (depending on the target_type param)
        :param permission: Permission (`Read`, `Write`, or `Share`) to be shared for a template. It can be Read, Write or Share.
        :return: Returns a SharePermission for the specific template if successful, or DtlError if failed
        """

        url = f"{self.service_uri}/templates/{str(id)}/shares?targetType={str(target_type.value)}&targetId={str(target_id)}&permission={str(permission.value)}"
        rsp = self.http_client.execute_authed_request(url, HttpMethod.POST)

        if isinstance(rsp, DtlError):
            return rsp
        elif rsp.status_code == 200:
            return SharePermission(object_type=ObjectType.PipelineTemplate).from_payload(rsp.json())
        else:
            return DtlError(rsp.text)

    def unshare(self, id: UUID, target_id: UUID, target_type: Scope, permission: Permission) -> Union[
        UnsharePermission, DtlError]:
        """
        Unshare a pipeline template.
        :param id: id of the template that is to be unshared
        :param target_type: Scope (`User`, `Group` or `Organization`) with whom you want to share the template. It can be User, Group or Organization.
        :param target_id: id of the User, Group or Organization you want to unshare with (depending on the target_type param)
        :param permission: Permission (`Read`, `Write`, or `Share`) to be unshared for a template. It can be Read, Write or Share.
        :return: Returns a UnsharePermission for the specific template if successful, or DtlError if failed
        """

        url = f"{self.service_uri}/templates/{str(id)}/shares?targetType={str(target_type.value)}&targetId={str(target_id)}&permission={str(permission.value)}"
        rsp = self.http_client.execute_authed_request(url, HttpMethod.DELETE)

        if isinstance(rsp, DtlError):
            return rsp
        elif rsp.status_code == 200:
            return UnsharePermission(object_type=ObjectType.PipelineTemplate).from_payload(rsp.json())
        else:
            return DtlError(rsp.text)

    def validate(self, id: UUID, dummy_input_map: Dict[str, str]) -> Union[bool, DtlError]:
        """
        Validate if an existing pipeline template compiles, with a supplied map of values.
        :param id: id of the template that is to be validated
        :param dummy_input_map: an example of data that will be used to
        validate the existing template
        :return: Returns True if template is valid, or DtlError if the validation was unsuccessful
        """
        url = f"{self.service_uri}/templates/{str(id)}/validate"
        rsp = self.http_client.execute_authed_request(url,
                                                      HttpMethod.POST,
                                                      dummy_input_map)

        if isinstance(rsp, DtlError):
            return rsp
        elif rsp.status_code == 200:
            return True
        else:
            return DtlError(rsp.text)

    def generate_pipeline(self,
                          pipeline_template_id: Union[str, UUID],
                          pipeline_name: str,
                          input_map: Dict[str, str]) -> Union[Pipeline, DtlError]:
        """
        Generate a pipeline by providing the missing inputs and their
        mapping to an existing pipeline template

        The correct number and mapping of inputs must be provided to the
        template—i.e. for a template with a {{local_path}}, a mapping of
        {"local_path" : "/data/source.csv"} would be appropriate

        :param pipeline_template_id: id of the pipeline template
        :param pipeline_name: name of the pipeline to be generated
        :param input_map: a mapping of inputs to be slotted into the template,
        following moustache distinction <https://mustache.github.io/mustache.5.html>
        :return: the newly created Pipeline if successful, or DtlError if
        failed
        """
        url = f"{self.service_uri}/templates/{str(pipeline_template_id)}/pipeline"
        payload = {
            "name": pipeline_name,
            "inputMap": input_map
        }
        rsp = self.http_client.make_authed_request(url,
                                                   HttpMethod.POST,
                                                   payload)

        if isinstance(rsp, DtlError):
            return DtlError(f"Encountered an error while generating pipeline from template: {rsp.message}")

        return Pipeline._from_payload(rsp)

    def run(self,
            pipeline_template_id: Union[str, UUID],
            pipeline_name: str,
            input_map: Dict[str, str]) -> Union[DtlError, Job]:
        """
        Run a pipeline completed by providing the missing inputs and their
        mapping to an existing pipeline template

        The correct number and mapping of inputs must be provided to the
        template—i.e. for a template with a {{local_path}}, a mapping of
        {"local_path" : "/data/source.csv"} would be appropriate

        :param pipeline_template_id: id of the pipeline template
        :param pipeline_name: name of the pipeline to be run
        :param input_map: a mapping of inputs to be slotted into the template,
        following moustache distinction <https://mustache.github.io/mustache.5.html>
        :return: a Job generated by running the completed pipeline if
        successful, or DtlError if failed
        """

        url = f"{self.service_uri}/templates/{str(pipeline_template_id)}/run"
        payload = {
            "name": pipeline_name,
            "inputMap": input_map
        }

        rsp = self.http_client.make_authed_request(url,
                                                   HttpMethod.POST,
                                                   payload)

        if isinstance(rsp, DtlError):
            return DtlError(f"Encountered an error while running pipeline from template: {rsp.message}")

        jobs = _parse_list(_job_from_payload)(rsp)
        if len(jobs) == 0:
            return DtlError("Encountered an issue! There should be at least one job!")
        else:
            return jobs[0]

    def schedule(self,
                 pipeline_template_id: Union[str, UUID],
                 pipeline_name: str,
                 input_map: Dict[str, str],
                 run_date: datetime,
                 frequency: Optional[timedelta] = None,
                 abort_if_overlong: bool = False) -> Union[DtlError, Job]:

        """
        Schedule a template to generate jobs, optionally following a
        recurrent frequency

        :param pipeline_template_id: id of the pipeline template
        :param pipeline_name: name of the pipeline to be run
        :param input_map: a mapping of inputs to be slotted into the
        template, following moustache distinction
        :param run_date: localized date at which to schedule the job
        :param frequency: frequency of subsequent re-runnings, via the
        timedelta class
        :param abort_if_overlong: defaults to run a subsequent running if
        a current job is still running
        :return: First scheduled job of the pipeline’s running if
        successful, or DtlError if failed
        """

        url = f"{self.service_uri}/templates/{str(pipeline_template_id)}/schedule"

        payload = {
            "name": pipeline_name,
            "runDate": run_date.astimezone(UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
            "inputMap": input_map,
            "abortIfOverlong": abort_if_overlong
        }

        if frequency is not None:
            payload.update({
                "intervalInSec": frequency.seconds
            })

        rsp = self.http_client.make_authed_request(url,
                                                   HttpMethod.POST,
                                                   payload)

        if isinstance(rsp, DtlError):
            return DtlError(f"Encountered an error while running pipeline from template: {rsp.message}")

        jobs = _parse_list(_job_from_payload)(rsp)
        if len(jobs) == 0:
            return DtlError("Encountered an issue! There should be at least one job!")
        else:
            return jobs[0]



from uuid import UUID
from typing import List, Union, Optional

from datalogue.errors import DtlError


class Organization:
    def __init__(self, org_id: UUID, name: str):
        self.id = org_id
        self.name = name

    def __eq__(self, other: 'Organization'):
        if isinstance(self, other.__class__):
            return self.id == other.id and self.name == other.name
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}(id: {self.id}, name: {self.name!r})'


def _organization_from_payload(json: dict) -> Union[DtlError, Organization]:
    org_id = json.get("id")
    if org_id is None:
        return DtlError("Organization object should have an 'id' property")
    else:
        try:
            org_id = UUID(org_id)
        except ValueError:
            return DtlError("'id' field was not a proper uuid")

    name = json.get("name")
    if name is None:
        return DtlError("Organization object should have a 'name' property")

    return Organization(org_id, name)


class Group:
    def __init__(self, group_id: UUID, org_id: UUID, name: str):
        self.id = group_id
        self.name = name
        self.org_id = org_id

    def __eq__(self, other: 'Group'):
        if isinstance(self, other.__class__):
            return self.id == other.id and self.name == other.name and self.org_id == other.org_id
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}(id: {self.id}, name: {self.name!r}, org_id: {self.org_id!r})'


def _group_from_payload(json: dict) -> Union[DtlError, Group]:
    group_id = json.get("id")
    if group_id is None:
        return DtlError("Group object should have an 'id' property")
    else:
        try:
            group_id = UUID(group_id)
        except ValueError:
            return DtlError("'id' field was not a proper uuid")

    name = json.get("name")
    if name is None:
        return DtlError("Group object should have a 'name' property")

    org_id = json.get("orgId")
    if org_id is None:
        return DtlError("Group object should have a 'org_id' property")
    else:
        try:
            org_id = UUID(org_id)
        except ValueError:
            return DtlError("'orgId' field was not a proper uuid")

    return Group(group_id, org_id, name)


class CreateUserRequest:
    """
    Definition of request needed to create user
    """
    def __init__(self,  email: str, first_name: str, last_name: Optional[str]):
        """
        constructor of CreateUserRequest
        :param email: email of the user
        :param first_name: first name of the user
        :param last_name: last name of the user
        """
        self.email = email
        self.first_name = first_name
        self.last_name = last_name

    def as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        return {
            'email': self.email,
            'firstName': self.first_name,
            'lastName': self.last_name,
        }


class User:
    """
    Definition of user
    """
    def __init__(self, user_id: UUID, first_name: Optional[str], last_name: Optional[str], email: str, organization_ids: Union[None,List[str]], group_ids: Union[None,List[str]]):
        """
        User constructor
        :param user_id: UUID of the user
        :param first_name: user first name
        :param last_name: user last name
        :param email: user email
        :param organization_ids: ids of organizations which users belongs to
        :param group_ids: ids of groups which users belongs to
        """
        self.id = user_id
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.organization_ids = organization_ids
        self.group_ids = group_ids

    def __eq__(self, other: 'User'):
        if isinstance(self, other.__class__):
            return self.id == other.id and self.first_name == other.first_name and self.last_name == other.last_name and self.email == other.email
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}(id: {self.id}, first_name: {self.first_name!r}, last_name: {self.last_name!r}, email: {self.email!r}, organization_ids: {self.organization_ids!r}, group_ids: {self.group_ids!r})'


def _user_from_payload(json: dict) -> Union[DtlError, User]:
    user_id = json.get("user").get("id")
    if user_id is None:
        return DtlError("User object should have an 'id' property")
    else:
        try:
            user_id = UUID(user_id)
        except ValueError:
            return DtlError("'id' field was not a proper uuid")

    first_name = json.get("user").get("firstName")
    last_name = json.get("user").get("lastName")

    email = json.get("user").get("email")
    if email is None:
        return DtlError("User object should have a 'email' property")

    organization_ids = json.get("organizationsIds")
    # if organization_ids is None:
    #     return DtlError("User object should have a 'organizationsIds' property")

    group_ids = json.get("groupsIds")
    if group_ids is None:
        return DtlError("User object should have a 'groupsIds' property")

    return User(user_id, first_name, last_name, email, organization_ids, group_ids)


def _users_from_payload(json: dict) -> Union[DtlError, User]:
    user_id = json.get("id")
    if user_id is None:
        return DtlError("User object should have an 'id' property")
    else:
        try:
            user_id = UUID(user_id)
        except ValueError:
            return DtlError("'id' field was not a proper uuid")

    first_name = json.get("firstName")
    last_name = json.get("lastName")

    email = json.get("email")
    if email is None:
        return DtlError("User object should have a 'email' property")

    group_ids = json.get("groupIds")

    return User(user_id, first_name, last_name, email, None, group_ids)


class Domain:
    def __init__(self, org_id: UUID, domain: str):
        self.org_id = org_id
        self.domain = domain

    def __eq__(self, other: 'Domain'):
        if isinstance(self, other.__class__):
            return self.org_id == other.org_id and self.domain == other.domain
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}(org_id: {self.org_id}, domain: {self.domain!r})'

    def __hash__(self):
        return hash(self.org_id) ^ hash(self.domain)

def _domain_from_payload(json: dict) -> Union[DtlError, Domain]:
    org_id = json.get("orgId")
    if org_id is None:
        return DtlError("Domain object should have an 'orgId' property")
    else:
        try:
            org_id = UUID(org_id)
        except ValueError:
            return DtlError("'id' field was not a proper uuid")

    domain = json.get("domain")
    if domain is None:
        return DtlError("Domain object should have a 'domain' property")

    return Domain(org_id, domain)


class TokenCreateResult:
    def __init__(self, id: UUID, token: str, checksum: str):
        self.id = id
        self.token = token
        self.checksum = checksum

    def __repr__(self):
        return f'{self.__class__.__name__}(id: {self.id}, token: {self.token!r}, checksum: {self.checksum!r})'


    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'TokenCreateResult']:
        id = json.get("id")
        if id is None:
            return DtlError("TokenCreateResult should have a 'id' property")

        token = json.get("token")
        if token is None:
            return DtlError("TokenCreateResult should have a 'token' property")

        checksum = json.get("checksum")
        if checksum is None:
            return DtlError("TokenCreateResult should have a 'checksum' property")
        return TokenCreateResult(id, token, checksum)


class PublicToken:

    def __init__(self, id: UUID, user_id: UUID, name: str, organization_id: UUID, checksum: str):
        self.id = id
        self.user_id = user_id
        self.name = name
        self.organization_id = organization_id
        self.checksum = checksum
    
    def __repr__(self):
        return (f'{self.__class__.__name__}(\n'
                f'id: {self.id!r},\n'
                f'user_id: UUID({self.user_id!r}),\n'
                f'name: {self.name!r},\n'
                f'organization_id: UUID({self.organization_id!r})\n'
                f'checksum: {self.checksum!r}\n'
                f')')

    @staticmethod
    def _from_payload(payload: dict) -> Union[DtlError, 'PublicToken']:
        id = payload.get("id")
        if id is None:
            return DtlError("TokenCreateResult should have a 'id' property")

        user_id = payload.get("userId")
        if user_id is None:
            return DtlError("TokenCreateResult should have a 'userId' property")

        name = payload.get("name")
        if name is None:
            return DtlError("TokenCreateResult should have a 'name' property")

        organization_id = payload.get("organizationId")
        if organization_id is None:
            return DtlError("TokenCreateResult should have a 'organizationId' property")

        checksum = payload.get("checksum")
        if checksum is None:
            return DtlError("TokenCreateResult should have a 'checksum' property")
        
        return PublicToken(id, user_id, name, organization_id, checksum)



import os
from codecs import open

from setuptools import find_packages, setup

dtl_version = {}
with open("./datalogue/version.py") as fp:
    exec(fp.read(), dtl_version)

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

dependencies = []

with open(os.path.join(ROOT_DIR, "./requirements.txt")) as fp:
    dependencies = fp.read().splitlines()


with open(os.path.join(ROOT_DIR, "./LICENSE"), "r") as license_file:
    imported_license = license_file.read()
    setup(
        name="datalogue",
        version=dtl_version["__version__"],
        author="Nicolas Joseph",
        author_email="nic@datalogue.io",
        license=imported_license,
        description="SDK to interact with the datalogue platform",
        long_description="",
        long_description_content_type="text/markdown",
        url="https://github.com/datalogue/dtl-python-sdk",
        packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
        classifiers=[
            "Programming Language :: Python :: 3",
            'Programming Language :: Python :: 3.6',
            "Operating System :: OS Independent"
        ],
        python_requires=">=3.6",
        install_requires=['requests', 'python-dateutil', 'validators', 'pytest==3.6.3', 'numpy', 'pyyaml', 'pyarrow', 'pandas', 'pbkdf2'],
        setup_requires=['pytest-runner'],
        tests_require=['pytest==3.6.3', 'pytest-cov==2.6.0']
)



import os
import json
import yaml
import numpy as np
from glob import glob
from typing import List, Union
import pandas as pd
from collections import namedtuple
import datetime

from datalogue.dtl_utils import _parse_list, _parse_string_list, SerializableStringEnum
from datalogue.models.model import TrainingState
from datalogue.errors import DtlError, _enum_parse_error


class ArgoMetrics:
    def __init__(self, argo_work_dir: str, experiment_name: str, class_map_path: str):
        if not os.path.isdir(argo_work_dir):
            raise DtlError(
                f"{argo_work_dir} is not a directory. Path to the template's work directory(i.e. nightshade/work/cbc"
            )
        self.argo_work_dir = argo_work_dir
        self.experiment_name = experiment_name
        self.class_map = self._load_class_map(class_map_path)
        self.training_files = self._fetch_training_files()
        self.test_file = self._fetch_test_file()
        self.payload = self.create_payload()

    def _fetch_training_files(self) -> List[str]:
        tfs = sorted(glob(os.path.join(self.argo_work_dir, self.experiment_name, "control",
                                       "*-train-completed", "epoch-*-experiment-tracking.json")))
        if len(tfs) == 0:
            raise DtlError(
                f"No Training Files found in {self.argo_work_dir}, experiment: {self.experiment_name}")
        return tfs

    def _fetch_test_file(self) -> str:
        tfs = sorted(glob(os.path.join(self.argo_work_dir, self.experiment_name, "control",
                                       "*-test-completed", "test-results.json")))
        if len(tfs) == 0:
            raise DtlError(
                f"No Test Files found in {self.argo_work_dir}, experiment: {self.experiment_name}")
        return tfs[-1]

    @staticmethod
    def _load_class_map(class_map_path: str) -> dict:
        if not os.path.isfile(class_map_path):
            raise DtlError(
                f"{class_map_path} is not a file. Please provide a Json or Yaml file!"
            )
        with open(class_map_path, 'r', encoding='utf-8') as f:
            if class_map_path.endswith('json'):
                return json.load(f)
            elif class_map_path.endswith('yaml'):
                return yaml.load(f)
            raise DtlError(
                f"{class_map_path} is not Json or Yaml. Unable to Load Class Map")

    def get_metrics(self, conf_mat: List[List], phase_name: str, class_names: List[str]) -> dict:
        rev_class_map = {v: k for k, v in self.class_map.items()}
        fp = conf_mat.sum(axis=0) - np.diag(conf_mat)
        fn = conf_mat.sum(axis=1) - np.diag(conf_mat)
        tp = np.diag(conf_mat)
        tn = conf_mat.sum() - (fp + fn + tp)
        epsilon = 1e-15  # So that we cannot get division by 0 errors
        # Sensitivity, hit rate, recall, or true positive rate
        tpr = tp / (tp + fn + epsilon)
        fpr = fp / (fp + tn + epsilon)
        recall = {rev_class_map[i]: float(tpr[i]) for i in range(len(tpr))}
        # Precision or positive predictive value
        ppv = tp / (tp + fp + epsilon)
        precision = {rev_class_map[i]: float(ppv[i]) for i in range(len(ppv))}
        # f1 Score
        f1 = 2 * tpr * ppv / (tpr + ppv + epsilon)
        f1_score = {rev_class_map[i]: float(f1[i]) for i in range(len(f1))}

        evaluation = {'f1': f1_score,
                      'recall': recall,
                      'precision': precision,
                      'fpr': np.mean(fpr),
                      'tpr': np.mean(tpr)
                      }

        return evaluation

    def create_payload(self) -> List[dict]:
        payload = []
        for file in self.training_files:
            with open(file, 'r') as f:
                data = json.load(f)
            M = {"epoch": data["epoch"], 'metrics': {}}
            for phase in ["training", "validation"]:
                M['metrics'][phase] = {
                    "loss": data["metrics"][phase]["loss"],
                    "acc": data["metrics"][phase]["acc"],
                    "avg_precision_loss": data["metrics"][phase]["avg_precision_loss"],
                    "precision":  data["metrics"][phase]["precision"],
                    "recall":  data["metrics"][phase]["recall"],
                    "f1": data["metrics"][phase]["f1"],
                    "tpr": data["metrics"][phase]["tpr"] if "tpr" in data["metrics"][phase] else None,
                    "fpr": data["metrics"][phase]["fpr"] if "fpr" in data["metrics"][phase] else None,
                    "confusion_matrix": None if "training" == phase else data["metrics"][phase]["confusion_matrix"]
                }
                if "validation" == phase:
                    conf_mat = np.array(
                        data["metrics"][phase]["confusion_matrix"]["matrix"])
                    evaluation = self.get_metrics(
                        conf_mat, phase_name=f"{self.experiment_name}_VAL_epoch_{data['epoch']}_", class_names=data["metrics"][phase]["confusion_matrix"]["index"])
                    M['metrics'][phase].update(evaluation)
            payload.append(M)

        with open(self.test_file, 'r') as f:
            test_data = json.load(f)

        M = {"epoch": len(self.training_files), 'metrics': {}}
        for phase in ["training", "validation"]:
            M['metrics'][phase] = {
                "loss": test_data["metrics"]["test"]["loss"],
                "acc": test_data["metrics"]["test"]["acc"],
                "avg_precision_loss": test_data["metrics"]["test"]["avg_precision_loss"],
                "precision": test_data["metrics"]["test"]["precision"],
                "recall": test_data["metrics"]["test"]["recall"],
                "f1": test_data["metrics"]["test"]["f1"],
                "confusion_matrix": test_data["metrics"]["test"]["confusion_matrix"]
            }

            conf_mat = np.array(test_data["metrics"]["test"]
                                ["confusion_matrix"]["matrix"])
            evaluation = self.get_metrics(conf_mat,
                                          phase_name=f"{self.experiment_name}_test_",
                                          class_names=test_data["metrics"]["test"]["confusion_matrix"]["index"])
            M['metrics'][phase].update(evaluation)
        payload.append(M)
        return payload


from uuid import UUID

from datalogue.errors import DtlError, _enum_parse_error, _invalid_parameter_error
from datalogue.dtl_utils import _parse_string_list, SerializableStringEnum
import abc
from typing import Union, List, Optional


class Transformation(abc.ABC):
    type_field = "type"

    def __init__(self, transformation_type: str):
        self.type = transformation_type
        super().__init__()

    def _base_payload(self) -> dict:
        return dict([(Transformation.type_field, self.type)])

    @abc.abstractmethod
    def _as_payload(self) -> dict:
        """
        Represents the transformation in its dictionary construction

        :return:
        """


class RegexTransformation(Transformation):
    def __init__(self, regex: Optional[str], regex_id: Optional[UUID], regex_transformation_type: str):
        Transformation.__init__(self, regex_transformation_type)
        if regex is None and regex_id is None:
            raise _invalid_parameter_error("neither regex nor regex_id is/are not defined!", regex_transformation_type)
        if regex is not None and regex_id is not None:
            raise _invalid_parameter_error("both regex and regex_id are defined! Please provide only either.",
                                           regex_transformation_type)


class DataType(SerializableStringEnum):
    """
    Data Types that can be specified in the pipeline
    """
    String = "String"
    Boolean = "Boolean"
    Integer = "Integer"
    DateTime = "DateTime"
    Double = "Double"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("data type", s)

    @staticmethod
    def from_str(string: str) -> Union[DtlError, 'DataType']:
        return SerializableStringEnum.from_str(DataType)(string)


def _array_from_dict(json: dict, transformation_type: str, array_field_key: str) -> Union[DtlError, List[str]]:
    """
    Generic method to extract an array from a json transformation dictionary

    :param json: dictionary built from json
    :param transformation_type: type of the transformation
    :param array_field_key: key at which the array is
    :return:
    """
    if json.get(Transformation.type_field) != transformation_type:
        return DtlError("Dictionary input is not of type %s" % transformation_type)

    array_field = json.get(array_field_key)
    if array_field is None:
        return DtlError("'%s' is missing from the json transformation" % array_field_key)

    array = _parse_string_list(array_field)

    if isinstance(array, DtlError):
        return array

    return array


class ByClass:
    def __init__(
            self,
            class_id: Union[str, UUID]
    ):
        self.class_id = class_id
        """
        Targets data classified as the specified class
        :param class_id: id of the class with which to specify as a transformation target
        """

    def __repr__(self):
        return ('ByClass('
                f'class_id: {self.class_id!r})')



class ByFieldName:
    def __init__(
            self,
            field_name: List[str]
    ):
        self.field_name = field_name
        """
        Targets data residing under the specified field name
    
        :param field_name: name of the field to specify as a transformation target
        """

    def __repr__(self):
        return ('ByFieldName('
                f'field_name: {self.field_name!r})')
